/*
 * RoombaBehave1
 * --------------
 * Implement the RoombaComm BumpTurn program in Arduino
 * A simple behavioural program that controls a Roomba
 * using SCI.
 * 
 * Arduino pin 4 (RX) is connected to Roomba TXD
 * Arduino pin 3 (TX) is connected to Roomba RXD
 * Arduino pin 2      is conencted to Roomba DD
 * 
 * Needs Arduino 1.0 or better IDE.
 * Copyright (c) Dennis Clark - Free to use by anyone.
 * Mr. Roboto April 2013
 */

#include <SoftwareSerial.h>

int rxPin = 4;
int txPin = 3;
int ddPin = 2;
int ledPin = 13;

/*
*  Define number of sensor bytes and the 
*  byte array.
*/
#define BYTES 10
uint8_t sensorbytes[BYTES];

SoftwareSerial mySerial(rxPin,txPin);  // Roomba comm port
/*
*  Define a bunch of sensor inspection macros.
*/
#define mBumpRight (uint8_t)(sensorbytes[0] & 0x01)
#define mBumpLeft  (uint8_t)(sensorbytes[0] & 0x02)
#define mBumpers (uint8_t)(sensorbytes[0] & 0x03)
#define RIGHT (uint8_t)1
#define LEFT (uint8_t)2
#define FRONT (uint8_t)3
#define mSeeWall  (sensorbytes[1])
#define mCliffLeft (sensorbytes[2]);
#define mCliffFrontLeft (sensorbytes[3])
#define mCliffFrontRight (sensorbytes[4])
#define mCliffRight (sensorbytes[5])
#define mVirtWall (sensorbytes[6])

/*
*  Define our behavioral functions and priorities
*/
typedef enum 
{
  idle = 0,    // do nothing
  wander = 1,  // mostly just drive straight
  avoid,       // use virtual wall
  escape,      // use bump sensors
  panic        // use cliff sensors
} eBehavior_t;

eBehavior_t priority = idle;

void setup() {
  uint8_t i;
  
  pinMode(ddPin,  OUTPUT);   // DD "wakeup" pin is an output
  pinMode(ledPin, OUTPUT);   // LED pin is an output
  
  /*
  *  Set up the software serial port to talk to the Roomba.
  */
  mySerial.begin(57600);

  /*
  *  Set up the hardware serial port for debugging.  
  */
  Serial.begin(57600);
  
  Serial.println("Wake up Roomba\n");
  digitalWrite(ledPin, HIGH); // say we're alive
  
  // Wake the Roomba up
  digitalWrite(ddPin, HIGH);
  digitalWrite(ddPin, LOW);
  delay(500);
  digitalWrite(ddPin, HIGH);
  
  /*
  * Grab whatever interesting stuff the Roomba transmits
  * and send it to debug port.
  */
  while(mySerial.available())
  {
    i = mySerial.read();
    Serial.write(i);
  }
  Serial.println();
  delay(2000);
  
  /*
  *  Set up Roomba to receive commands in 
  *  "Safe" mode.
  */
  Serial.println("set up ROI");
  mySerial.write((uint8_t)128);    // START
  delay(50);
  mySerial.write((uint8_t)130);    // CONTROL
  delay(50);
  digitalWrite(ledPin, LOW);       // say we've finished setup  

  mySerial.write((uint8_t)140);    // create a song
  mySerial.write((uint8_t)0);      // song 0
  mySerial.write((uint8_t)2);      // two notes
  mySerial.write((uint8_t)36);
  mySerial.write((uint8_t)7);
  mySerial.write((uint8_t)60);
  mySerial.write((uint8_t)7);
  
}

/*
*  The nice thing about state machines is that they allow your
*  main loop to be really simple.
*/
void loop(void) 
{
  UpdateSensors();
  DoWall();
  DoWander();
  DoAvoid();
  DoEscape();
  DoPanic();
}

/*
* Behavior definitions
*/

void DoWander(void)
// Define a wander algorithm, or just go straight.
{
  if (priority < wander) 
  {
    Serial.println("Wander");
    priority = wander;
    GoForward();
  }
}

void DoWall(void)
{
  if (sensorbytes[1]) {
    Serial.println("WALL!");
    GoBrush(1);
    GoLEDs(1);
  }
  else {
    GoBrush(0);
    GoLEDs(0);
  }
}

void DoEscape(void)
// Define a bumper response escape behavior
{
  static uint32_t ticker = 0;
  static uint8_t state = 0;
  static uint8_t dir = 0;
  
  if ((mBumpers == 0) && (state ==0))
  // Don't do anything if nothing to do
  {
    return;
  }
  if (priority <= escape)
  {
    priority = escape;
    switch(state) 
    {
      case 0: // record bumper, backup
        dir = mBumpers;
        ticker = millis() + 1000;  // 1 second
        GoBackward();
        GoSong();
        state = 1;
        Serial.print("Bumper dir: ");
        Serial.println(dir,HEX);
        break;
        
      case 1:           // wait for backup done
        if (ticker < millis())
        {
          state = 2;
          Serial.println("Do turn");
        }
        break;
        
      case 2:           // choose turn direction 
        if (dir == RIGHT)
        {
          Serial.println("spin left");
          SpinLeft();
        }
        else if (dir == LEFT)
        {
          Serial.println("spin right");
          SpinRight();
        }
        else 
        {
          Serial.println("ahead");
          SpinRight();
        }
         ticker = millis() + 1000;
         state = 3;
        break;
        
      case 3:           // wait for turn done
        if (ticker < millis())
        {
          state = 0;
          dir = 0;
          priority = idle;
          Serial.println("All done");
        }
        break;
    }
  }
}

void DoAvoid(void)
// Define an object avoidance, mostly from virtual wall
{
  static uint32_t ticker = 0;
  static uint8_t state = 0;
  
}

void DoPanic(void)
// Cliff sensor or wheel drop detected
{
  static uint32_t ticker = 0;
  static uint8_t state = 0;
  
}

/*
* Basic SCI movement, sound and sensor commands
*/
void GoSong(void) {
  mySerial.write((uint8_t)141);  // play a song
  mySerial.write((uint8_t)0);
  delay(5);
}

void GoBrush(uint8_t on) {
  // Turn side brush on or off
  if (on)
  {
    mySerial.write((uint8_t)138);
    mySerial.write((uint8_t)0x01);
  }
  else 
  {
    mySerial.write((uint8_t)138);
    mySerial.write((uint8_t)0x00);
  }
  delay(5);
}

void GoLEDs(uint8_t on) {
  // Turns on/off Spot, Clean, Dirt and MAX LEDs, if there.
  if (on)
  {
    mySerial.write((uint8_t)139);
    mySerial.write((uint8_t)15);
    mySerial.write((uint8_t)0);
    mySerial.write((uint8_t)0);
  }
  else 
  {
    mySerial.write((uint8_t)139);
    mySerial.write((uint8_t)0);
    mySerial.write((uint8_t)0);
    mySerial.write((uint8_t)0);
  }
  delay(5);
}

void GoForward() {
  mySerial.write((uint8_t)137);   // DRIVE
  mySerial.write((uint8_t)0x00);   // 0x00c8 == 200
  mySerial.write((uint8_t)0xc8);
  mySerial.write((uint8_t)0x80);
  mySerial.write((uint8_t)0x00);
  delay(5);
}

void GoBackward() {
  mySerial.write((uint8_t)137);   // DRIVE
  mySerial.write((uint8_t)0xff);   // 0xff38 == -200
  mySerial.write((uint8_t)0x38);
  mySerial.write((uint8_t)0x80);
  mySerial.write((uint8_t)0x00);
  delay(5);
}

void SpinLeft() {
  mySerial.write((uint8_t)137);   // DRIVE
  mySerial.write((uint8_t)0x00);   // 0x00c8 == 200
  mySerial.write((uint8_t)0xc8);
  mySerial.write((uint8_t)0x00);
  mySerial.write((uint8_t)0x01);   // 0x0001 == spin left
  delay(5);
}

void SpinRight() {
  mySerial.write((uint8_t)137);   // DRIVE
  mySerial.write((uint8_t)0x00);   // 0x00c8 == 200
  mySerial.write((uint8_t)0xc8);
  mySerial.write((uint8_t)0xff);
  mySerial.write((uint8_t)0xff);   // 0xffff == -1 == spin right
  delay(5);
}
void UpdateSensors() {
  int8_t  i = 0;
  mySerial.write((uint8_t)142);
  mySerial.write((uint8_t)1);  // sensor packet 1, 10 bytes

  while(!mySerial.available());  // wait for data to appear.
  
  while(mySerial.available()    // get 10 bytes of sensor data
  &&   (i < BYTES) ) 
  {
    int c = mySerial.read();
    if( c==-1 ) {
      for( int i=0; i<5; i ++ ) {   // say we had an error via the LED
        digitalWrite(ledPin, HIGH); 
        delay(50);
        digitalWrite(ledPin, LOW);  
        delay(50);
      }
    }
    sensorbytes[i++] = c;
  }      
  delay(5);
  //Serial.print("Got ");Serial.print(i);Serial.println(" bytes");
  //i = 0;
  //while (i++ < BYTES)
  //{
  //  Serial.print(i,HEX); Serial.print(" ");
  //}
  //Serial.println();
}

    
